# Generated by Django 2.2.5 on 2020-04-28 15:03

import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion
import jsoneditor.fields.django_extensions_jsonfield


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='ImportJob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('options', jsoneditor.fields.django_extensions_jsonfield.JSONField(blank=True, default=dict, help_text='Detailed instructions for import if necessary', verbose_name='Options')),
                ('upload_file', models.FileField(storage=django.core.files.storage.FileSystemStorage(location=''), upload_to='uploads', verbose_name='Upload File')),
                ('model', models.ForeignKey(help_text='Specify a model to be imported', on_delete=django.db.models.deletion.CASCADE, related_name='imports', to='contenttypes.ContentType', verbose_name='Model')),
            ],
            options={
                'verbose_name': 'Import Job',
                'verbose_name_plural': 'Import Jobs',
            },
        ),
        migrations.CreateModel(
            name='ImportLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('imported_at', models.DateTimeField(auto_now=True, help_text='Timestamp when the import happens', verbose_name='Imported At')),
                ('import_log', models.TextField(default='', editable=False, help_text='Detailed log of the importing process', verbose_name='Import Log')),
                ('is_finished', models.BooleanField(default=False, editable=False, verbose_name='Is Finished')),
                ('job', models.ForeignKey(help_text='Job to which the log is related', on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='django_import.ImportJob', verbose_name='Job')),
            ],
            options={
                'verbose_name': 'Import Log',
                'verbose_name_plural': 'Import Logs',
            },
        ),
    ]



"""
Reflections are determined by the `reflections` section of the ImportJob instance options
attribute.

Every reflection is a key-value pair, where the key is a field name, while the value is
a reflection option which determines a way to reflect data row values to this field value.

The easiest reflection option is a string. If the reflection option is a string, it corresponds
to the reflection function name, and assumes that the data contains a column having the name
equal to the field name.

Alternatively, reflection option may have a form of the key-value section, having the following structure:

- `function` determines a reflection function name

- `parameters` determines a key-value section of additional reflection function parameters

A default reflection function is `direct`, it is used if the field is not listed in the
`reflections` section.

When the data row is got, all reflection functions are called, and returned values are collected.

The returned values may be used in two stages:

- create - an instance is created or updated using
    [`update_or_create()`](https://docs.djangoproject.com/en/2.2/ref/models/querysets/#update-or-create)
    function call

- update - every value collected for this stage is assigned to the correspondent
  instance attribute, and the instance is saved

Every reflection returns data for create and update stages separately.
"""
from functools import partial

from django.core.exceptions import FieldDoesNotExist
from django.utils.translation import ugettext_lazy as _


def _get_field(model, field_name):
    """Internal helper to get the model field"""
    try:
        return model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return None


def reflection_direct(context, model, field_name, data, log, column=None):
    """
`direct` reflection sends the column value to the instance create parameters

available parameters:

- column - column name where to find a value, field name by default
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    if column is None:
        column = field_name
    if column not in data:
        return {}, {}
    value = data.get(column)
    return {field_name: value}, {}


def reflection_update(context, model, field_name, data, log, column=None):
    """
`update` reflection sends the column value to the instance update parameters

available parameters:

- column - column name where to find a value, field name by default
    """
    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    update.update(create)
    return update, {}


def reflection_constant(context, model, field_name, data, log, value=None):
    """
`constant` reflection sets the field value to the constant

available parameters:

- value - value to be set
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    return {field_name: value}, {}


def reflection_avoid(*av, **kw):
    """
`avoid` reflection excludes the field from the import, to
avoid unnecessary import from the data column with the name
equal to the field name.

Any additional parameters are avoided also.
    """
    return {}, {}


def reflection_clean(context, model, field_name, data, log, column=None):
    """
`clean` reflection cleans the column value by the field clean function

available parameters:

- column - column name where to find a value, field name by default
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}

    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    if not create:
        return {}, {}
    value = field.clean(list(create.values())[0], model)
    return {field_name: value}, {}


def reflection_substr(context, model, field_name, data, log, column=None, start=0, length=None):
    """
`substr` reflection gets a substring from the data column

available parameters:

- column - column name where to find a value, field name by default
- start - start index in the string, 0 by default
- length - maximal number of characters to get from the value, max_length attribute
  of the field by default
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    if length is None:
        length = getattr(field, 'max_length', 0)
        if not length:
            log.warning(_('Can not determine length for the substr: %s'), field_name)
            return {}, {}
    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    if not create:
        return {}, {}
    value = list(create.values())[0]
    value = value[start: start + length]
    return {field_name: value}, {}


def reflection_replace(context, model, field_name, data, log, column=None, search=None, replace=None, count=None):
    """
`replace` reflection gets a value from the data column with `search` value replaced by the `replace` value.

available parameters:

- column - column name where to find a value, field name by default
- search - substring to be searched
- replace - value to replace
- count - number of occurences to replace, all by default
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    if search is None:
        log.warning(_('Set the search value: %s'), field_name)
        return {}, {}
    if replace is None:
        replace = ''
    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    if not create:
        return {}, {}
    value = list(create.values())[0]
    if count:
        value = value.replace(search, replace, count)
    else:
        value = value.replace(search, replace)
    return {field_name: value}, {}


def reflection_format(context, model, field_name, data, log, format='<format not set>'):
    """
`format` reflection creates the field value using %-style formatting from the whole data dict

available parameters:

- format - %-style format to create a value from the dict with columns
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    value = format % data
    return {field_name: value}, {}


def reflection_xformat(context, model, field_name, data, log, format='<format not set>'):
    """
`xformat` reflection creates the field value using {}-style formatting from the whole data dict

available parameters:

- format - {}-style format to create a value from the dict with columns
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    value = format.format(data)
    return {field_name: value}, {}


def reflection_enum(context, model, field_name, data, log, column=None, mapping={}):
    """
`enum` reflection uses a mapping from the column
to field values

available parameters:

- column - column name where to find a value, field name by default
- mapping - column-to-field value mapping (column value as a key)
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}

    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    if not create:
        return {}, {}
    value = list(create.values())[0]
    if value not in mapping:
        log.warning(_('Found a column value %s not in mapping: %s'), (value, field_name))
        return {}, {}
    value = mapping[value]
    return {field_name: value}, {}


def reflection_combine(context, model, field_name, data, log, reflections=[]):
    """
`combine` reflection applies several reflections sequentially,
getting output of the first and sending it as an input for
the second etc.

The output of the last is an output of the combine.

available parameters:

- reflections - a list of reflections to be applied
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}
    for r in reflections:
        if isinstance(r, str):
            r = {'function': r}
        function_name = r.get('function', None)
        if not function_name:
            log.warning(_('Function name is empty in combine: %s'), field_name)
            return {}, {}
        from django_import import reflections
        function = getattr(reflections, 'reflection_%s' % function_name, None)
        if not function:
            log.warning(_('Function name %s not found in combine: %s'), function_name, field_name)
            return {}, {}
        f = partial(function, **r.get('parameters', {}))
        c, u = f(context, model, field_name, data, log)
        data = {}
        data.update(c)
        data.update(u)
    return c, u


def reflection_lookup(context, model, field_name, data, log, column=None, lookup_field='pk'):
    """
`lookup` reflection is applied only to the foreign key or one-to-one
field reference field.

It lookups the column value in the `lookup_field` on the opposite side of
the reference, and gets the found instance, or None, if not found.

available parameters:

- column - column name where to find a value, field name by default
- lookup_field - field name of the opposide side model with the optional lookup suffix, 'pk' by default
    """
    field = _get_field(model, field_name)
    if not field:
        return {}, {}

    create, update = reflection_direct(context, model, field_name, data, log, column=column)
    if not create:
        return {}, {}
    value = list(create.values())[0]
    remote_model = field.remote_field.model
    value = remote_model.objects.filter(**{lookup_field: value}).last()
    return {field_name: value}, {}



from __future__ import unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class Config(AppConfig):
    verbose_name = _('Django Import')
    name = 'django_import'



from functools import partial

import pandas
from six import string_types

from django.db import DatabaseError, transaction
from django.utils.translation import ugettext_lazy as _

from . import reflections as reflect
from .config import get_options
from .reflector import Reflector


def run_import(import_log_id=None):
    """
    Evaluates importing process just after the ImportJob instance has been saved.

    It may be evaluated synchronously in the context of the WEB Application,
    or asynchronously in the context of the Celery Worker,
    depending on `async` option of the `DJANGO_IMPORT` variable in the settings file.
    """
    from django_import.models import ImportLog

    log = ImportLog.objects.filter(id=import_log_id).last()
    if not log:
        return

    job = log.job
    log.info(_('Trying to import %s'), job.upload_file)

    try:
        try_import(log)
    except Exception as ex:
        log.error(_('Unexpected error: %s'), ex)
    log.finish()


def try_import(log):
    job = log.job
    format_parameters = job.options.get('parameters', {})
    format = job.options.get('format', 'csv')
    mode = job.options.get('mode', 'rb')
    headers = job.options.get('headers', None)
    reflections = job.options.get('reflections', {})
    identity = job.options.get('identity', [])
    # TODO: file encoding? data encoding? leave as-is a while ...
    if mode not in ['rb', 'rt']:
        log.warning(_('Mode should be either rb (read binary), or rt (read text), got %s, ignored'), mode)
        mode = 'rb'

    read_function = getattr(pandas, 'read_%s' % format, None)
    if not read_function:
        log.warning(_('Read function not found, finished: read_%s'), format)

    params = {}
    params.update(**format_parameters)

    job.upload_file.open(mode)
    try:
        dataset = read_function(job.upload_file, **format_parameters)
    except Exception:
        job.upload_file.close()
        raise
    job.upload_file.close()
    log.info(
        _('Import file has been recognized, %s columns, %s rows: %s'),
        len(dataset.columns), len(dataset.index), job.upload_file
    )
    if headers:
        dataset = dataset.rename(columns=dict(zip([a for a in dataset.columns], headers)))
    if isinstance(dataset.columns, pandas.core.indexes.numeric.Int64Index):
        dataset = dataset.rename(columns=dict(zip([a for a in dataset.columns], ['%04d' % (a + 1) for a in dataset.columns])))
        log.info(_('Headers not found, replacing by sequential numbers: %s'), ', '.join(dataset.columns))
    model = job.model.model_class()
    convertors = {}
    reflector = Reflector()
    for f in model._meta.get_fields():
        reflection = reflections.get(f.name, 'direct')
        if isinstance(reflection, string_types):
            reflection = {
                'function': reflection
            }
        if not isinstance(reflection, dict):
            log.warning(_("The reflection is not formatted properly: %s"), reflection)
            continue
        reflection_function = getattr(reflect, 'reflection_%s' % reflection['function'], None)
        if not reflection_function:
            log.warning(_('Reflection function %s has not been registered, ignored'), reflection['function'])
            continue
        convertors[f.name] = {
            'function': reflection_function,
            'parameters': reflection.get('parameters', {})
        }
    cnt = 0
    skipped = 0
    rows_report = get_options()['rows_report']
    for ind, data in [(row[0], dict(zip(dataset.columns, row[1]))) for row in dataset.iterrows()]:
        create, update = {}, {}
        try:
            for field_name in convertors:
                f = partial(convertors[field_name]['function'], **convertors[field_name]['parameters'])
                c, u = f(reflector, model, field_name, data, log)
                create.update(c)
                update.update(u)
        except Exception as ex:
            log.warning(_("Error while importing data: %s"), ex)
            continue
        if not create and not update:
            if not cnt and skipped >= 2:
                log.warning(_("%s rows at the top have no reflected data, import interrupted"), skipped + 1)
                break
            else:
                log.warning(_("No any reflected data found, row skipped: %s"), ', '.join(['%s:%r' % (k, v) for k, v in data.items()]))
                skipped += 1
                continue
        try:
            with transaction.atomic():
                ident = dict([(k, v) for k, v in create.items() if k in identity])
                if ident:
                    instance, created = model.objects.update_or_create(defaults=create, **ident)
                else:
                    instance = model.objects.create(**create)
                if update:
                    for k in update:
                        setattr(instance, k, update[k])
                    instance.save()
        except DatabaseError as ex:
            log.error(_("Database error while importing data: create %r, update %r, %s"), create, update, ex)
            continue
        cnt += 1
        if cnt % rows_report == 0:
            log.info(_('... %s rows successfully imported ...'), cnt)
    log.info(_('Import has been finished, %s rows successfully imported'), cnt)



import markdown

from django.apps import apps
from django.contrib import admin
from django.contrib.admin import ModelAdmin, StackedInline
from django.contrib.contenttypes.models import ContentType
from django.db.models import F, Value
from django.db.models.functions import Concat
from django.utils.safestring import mark_safe

from .models import ImportJob, ImportLog, get_options


class ImportLogInline(StackedInline):
    readonly_fields = [
        'imported_at',
        'is_finished',
        'import_log_html',
    ]
    model = ImportLog
    extra = 0

    def has_add_permission(request, *av, **kw):
        return False


@admin.register(ImportJob)
class ImportJobAdmin(ModelAdmin):
    fields = [
        "model",
        "options",
        "upload_file",
        "extra_help"
    ]
    readonly_fields = [
        "extra_help"
    ]
    list_display = ["id", "model", "upload_file"]

    inlines = [
        ImportLogInline
    ]

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == 'model':
            queryset = ContentType.objects.all().annotate(full_name=Concat(F('app_label'), Value('.'), F('model')))
            options = get_options()
            if not options.get('models', []):
                models = ['%s.%s' % (m._meta.app_label, m._meta.model_name) for m in apps.get_models()]
                queryset = queryset.filter(
                    full_name__in=models
                ).exclude(
                    full_name__in=[x.lower() for x in options.get('except', [])]
                )
            else:
                queryset = queryset.filter(full_name__in=[m.lower() for m in options.get('models', [])])
            kwargs['queryset'] = queryset.order_by('model')
        return super(ImportJobAdmin, self).formfield_for_foreignkey(db_field, request, **kwargs)

    def extra_help(self, *av, **kw):
        try:
            from . import reflections
            help = '# Options\n%s\n# Reflections\n%s\n# Actual reflections list\n' % (ImportJob.__doc__, reflections.__doc__)
            for r in reflections.__dict__:
                func = getattr(reflections, r, None)
                if func and callable(func) and func.__name__.startswith('reflection_'):
                    help += '\n### %s\n%s' % (func.__name__.split('_', 1)[1], func.__doc__)
            help = markdown.markdown(help, extensions=['extra', 'smarty', 'sane_lists', 'attr_list'], output='html5')
            help = help.replace('<ul>', '<ul style="margin-left:20px;">').replace('<li>', '<li style="list-style-type: square;">')
        except Exception as ex:
            help = str(ex)
        return mark_safe(help)



from django.conf import settings


DJANGO_IMPORT = {
    'storage': {
        'class': 'django.core.files.storage.FileSystemStorage',
        'settings': {
            'location': '',
        },
        'upload_to': 'uploads',
    },
    'models': [],
    'except': [
        'contenttypes.ContentType',
        'admin.LogEntry',
        'auth.Permission',
        'sessions.Session',
        'django_import.ImportJob',
        'django_import.ImportLog',
    ],
    'sync': True,
    'rows_report': 1000,
}


def get_options():
    """
    Returns options from the settings.DJANGO_IMPORT,
    or default values
    """
    options = {}
    options.update(DJANGO_IMPORT)
    options.update(getattr(settings, 'DJANGO_IMPORT', {}))
    return options



from . import reflections


class Reflector(object):
    """
    Arbitrary context for the import process when evaluating reflections
    """


def register_reflection(name, func):
    """
    Register a new reflection function

    The reflection function should have the following signature:
    ```
    def reflection(context, model, field_name, data, log, **kwargs):
        ...
        return create, update
    ```
    where:
        - `context` is an instance of the `Reflector`
        - `model` is a model to be filled
        - `field_name` - is a name of the field to be filled
        - `data` - dictionary of all values got from the row
        - `log` - instance of the ImportLog model to be used to send logs if necessary
        - `kwargs` - additional parameters which are used from the job options.reflections.<field_name>.parameters chapter

    The returned values pair may be used in two stages:
        - create - an instance is created or updated using `update_or_create()` function call
        - update - every value collected for this stage is assigned to the correspondent
          instance attribute, and the instance is saved
    """
    setattr(reflections, name, func)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from django_import.version import __version__ as version


with open("README.rst", "r") as fp:
    description = fp.read() + "\n"

setup(
    name="django-import",
    version=version,
    description="Import to django models",
    long_description=description,
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        'Development Status :: 4 - Beta',
        "Framework :: Django",
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    ],
    keywords="CSV JSON TSV import django fixture",
    license='LGPL',
    packages=["django_import", "django_import.migrations"],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "django",
        "pandas",
        "django-jsoneditor",
        "django-extensions",
        "markdown",
        "six",
        "xlrd>=1.0.0"
    ],
)



import re

from jsoneditor.fields.django_extensions_jsonfield import JSONField

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.functional import LazyObject
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from .config import get_options


def get_file_storage_class():
    """
    Function returning a class as a file storage class.
    """
    options = get_options()
    return import_string(options['storage']['class'])


class Storage(LazyObject):
    """
    Lazy storage instance class
    to setup additional parameters from settings
    """
    def _setup(self):
        options = get_options()
        options = options['storage']['settings']
        storage_class = get_file_storage_class()
        self._wrapped = storage_class(**options)


class ImportJob(models.Model):
    """
Every time the ImportJob instance is saved, the importing process
is called, synchronously, or asynchronously, depending on the
`async` value of the `DJANGO_IMPORT` variable in the settings
file.

The `options` attribute has the complex structure to setup
various custom import options.

- `format` is a format to be imported. It selects one of pandas module functions to be
    used for import. Available values can be got from the pandas documentation as a list of `read_*`
    functions, like `table`, `csv`, `fwf`, `excel`, `json`, or any other reading function returning
    a `DataFrame`, read [pandas documentation](https://pandas.pydata.org/docs/reference/io.html)

- `parameters` section determines parameters to be sent as additional
    parameters to the reading function. Check [pandas documentation](https://pandas.pydata.org/docs/reference/io.html)
    to see what additional parameters may or should be sent there

- the `mode` determines file open mode when the file is got from the storage;
    the only two, `rb` (read binary) and `rt` (read text) modes are supported;
    *note* that some custom storages don't support proper mode changing options

- the `headers` determines a list of headers which should be assigned to
    data columns. Existent heades if present, are replaced. Length of the
    `headers` list should be equal to number of data columns

+ `identity` is a list of field column names which are used to determine identity
    of the database instance. It is used by the system in the call to
    [`update_or_create()`](https://docs.djangoproject.com/en/2.2/ref/models/querysets/#update-or-create) method

+ `reflections` determines customization in translation data to
    field values.
    """

    model = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name='imports',
        verbose_name=_("Model"),
        help_text=_("Specify a model to be imported"),
    )
    options = JSONField(
        blank=True,
        default=dict,
        verbose_name=_('Options'),
        help_text=_('Detailed instructions for import if necessary'),
    )
    upload_file = models.FileField(
        upload_to=get_options()['storage']['upload_to'], storage=Storage(),
        verbose_name=_('Upload File'),
    )

    def __str__(self):
        return '%(upload_file)s->%(model)s' % {
            'upload_file': self.upload_file,
            'model': self.model
        }

    class Meta:
        verbose_name = _('Import Job')
        verbose_name_plural = _('Import Jobs')

    def save(self, *av, **kw):
        super(ImportJob, self).save(*av, **kw)
        log = ImportLog.objects.create(job=self)
        log.info(_('Starting import for: %s'), self.upload_file)
        sync = get_options().get('sync', True)
        if not sync:
            from .tasks import run_import
            run_import.delay(log.id)
        else:
            from .import_task import run_import
            run_import(log.id)


class ImportLog(models.Model):
    job = models.ForeignKey(
        ImportJob, on_delete=models.CASCADE,
        related_name='logs',
        verbose_name=_('Job'),
        help_text=_('Job to which the log is related'),
    )
    imported_at = models.DateTimeField(
        auto_now=True,
        verbose_name=_('Imported At'),
        help_text=_('Timestamp when the import happens'),
    )
    import_log = models.TextField(
        editable=False,
        default='',
        verbose_name=_('Import Log'),
        help_text=_("Detailed log of the importing process"),
    )
    is_finished = models.BooleanField(
        editable=False,
        default=False,
        verbose_name=_('Is Finished'),
    )

    def import_log_html(self):
        return mark_safe(re.sub("\n", "<br/>", self.import_log))
    import_log_html.allow_tags = True
    import_log_html.short_description = _("Detailed Import Log")

    class Meta:
        verbose_name = _('Import Log')
        verbose_name_plural = _('Import Logs')

    def __str__(self):
        return '%(imported_at)s: %(job)s' % {
            'imported_at': self.imported_at,
            'job': self.job
        }

    def message(self, level, chapter, format, *av, **kw):
        now = timezone.now()
        values = av if av else kw
        try:
            message = '%s: [%s(%s)] %s' % (now, chapter, level, (format % values))
        except Exception as ex:
            try:
                r = '%r' % values
            except Exception:
                r = '<something strange>'
            try:
                f = '%r' % format
            except Exception:
                f = '<something strange>'
            self.error('Error formatting %s using %s: %s', r, f, ex)
        self.import_log += '\n%s' % message
        self.save()

    def debug(self, format, *av, **kw):
        return self.message(5, 'DEBUG', format, *av, **kw)

    def info(self, format, *av, **kw):
        return self.message(4, 'INFO', format, *av, **kw)

    def warning(self, format, *av, **kw):
        return self.message(3, 'WARNING', format, *av, **kw)

    def error(self, format, *av, **kw):
        return self.message(2, 'ERROR', format, *av, **kw)

    def critical(self, format, *av, **kw):
        return self.message(1, 'CRITICAL', format, *av, **kw)

    def finish(self):
        self.is_finished = True
        self.info(_('Finished'))
